SolarPressure

Struct SolarPressure 

pub struct SolarPressure {
    pub phi: f64,
    pub shadow_model: ShadowModel,
    pub estimate: bool,
}
Computation of solar radiation pressure is based on STK: http://help.agi.com/stk/index.htm#gator/eq-solar.htm .

:type shadow_bodies: listFrame :type almanac: Almanac :type flux_w_m2: float :type estimate: bool

Fields

phi: f64

solar flux at 1 AU, in W/m^2

shadow_model: ShadowModelestimate: bool

Set to true to estimate the coefficient of reflectivity

Implementations

Source

impl SolarPressure

pub fn default_flux_raw( shadow_bodies: Vec<Frame>, almanac: &Almanac, ) -> Result<Self, DynamicsError>

Will set the solar flux at 1 AU to: Phi = 1367.0

pub fn default_flux( shadow_body: Frame, almanac: &Almanac, ) -> Result<Arc<Self>, DynamicsError>

Accounts for the shadowing of only one body and will set the solar flux at 1 AU to: Phi = 1367.0

Examples found in repository?
nyx-core/examples/03_geo_analysis/stationkeeping.rs (line 93)
28fn main() -> Result<(), Box<dyn Error>> {
29    pel::init();
30    // Set up the dynamics like in the orbit raise.
31    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
32    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
33
34    // Define the GEO orbit, and we're just going to maintain it very tightly.
35    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
36    let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
37    println!("{orbit:x}");
38
39    let sc = Spacecraft::builder()
40        .orbit(orbit)
41        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0)) // 1000 kg of dry mass and prop, totalling 2.0 tons
42        .srp(SRPData::from_area(3.0 * 6.0)) // Assuming 1 kW/m^2 or 18 kW, giving a margin of 4.35 kW for on-propulsion consumption
43        .thruster(Thruster {
44            // "NEXT-STEP" row in Table 2
45            isp_s: 4435.0,
46            thrust_N: 0.472,
47        })
48        .mode(GuidanceMode::Thrust) // Start thrusting immediately.
49        .build();
50
51    // Set up the spacecraft dynamics like in the orbit raise example.
52
53    let prop_time = 30.0 * Unit::Day;
54
55    // Define the guidance law -- we're just using a Ruggiero controller as demonstrated in AAS-2004-5089.
56    let objectives = &[
57        Objective::within_tolerance(
58            StateParameter::Element(OrbitalElement::SemiMajorAxis),
59            42_165.0,
60            20.0,
61        ),
62        Objective::within_tolerance(
63            StateParameter::Element(OrbitalElement::Eccentricity),
64            0.001,
65            5e-5,
66        ),
67        Objective::within_tolerance(
68            StateParameter::Element(OrbitalElement::Inclination),
69            0.05,
70            1e-2,
71        ),
72    ];
73
74    let ruggiero_ctrl = Ruggiero::from_max_eclipse(objectives, sc, 0.2)?;
75    println!("{ruggiero_ctrl}");
76
77    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
78
79    let mut jgm3_meta = MetaFile {
80        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
81        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
82    };
83    jgm3_meta.process(true)?;
84
85    let harmonics = GravityField::new(GravityFieldData::from_cof(
86        &jgm3_meta.uri,
87        8,
88        8,
89        almanac.frame_info(IAU_EARTH_FRAME)?,
90    )?);
91    orbital_dyn.accel_models.push(harmonics);
92
93    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
94    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn)
95        .with_guidance_law(ruggiero_ctrl.clone());
96
97    println!("{sc_dynamics}");
98
99    // Finally, let's use the Monte Carlo framework built into Nyx to propagate spacecraft.
100
101    // Let's start by defining the dispersion.
102    // The MultivariateNormal structure allows us to define the dispersions in any of the orbital parameters, but these are applied directly in the Cartesian state space.
103    // Note that additional validation on the MVN is in progress -- https://github.com/nyx-space/nyx/issues/339.
104    let mc_rv = MvnSpacecraft::new(
105        sc,
106        vec![StateDispersion::zero_mean(
107            StateParameter::Element(OrbitalElement::SemiMajorAxis),
108            3.0,
109        )],
110    )?;
111
112    let my_mc = MonteCarlo::new(
113        sc, // Nominal state
114        mc_rv,
115        "03_geo_sk".to_string(), // Scenario name
116        None, // No specific seed specified, so one will be drawn from the computer's entropy.
117    );
118
119    // Build the propagator setup.
120    let setup = Propagator::rk89(
121        sc_dynamics.clone(),
122        IntegratorOptions::builder()
123            .min_step(10.0_f64.seconds())
124            .error_ctrl(ErrorControl::RSSCartesianStep)
125            .build(),
126    );
127
128    let num_runs = 25;
129    let rslts = my_mc.run_until_epoch(setup, almanac.clone(), sc.epoch() + prop_time, num_runs);
130
131    assert_eq!(rslts.runs.len(), num_runs);
132
133    rslts.to_parquet("03_geo_sk.parquet", ExportCfg::default())?;
134
135    Ok(())
136}
More examples
Hide additional examples
nyx-core/examples/03_geo_analysis/raise.rs (line 119)
27fn main() -> Result<(), Box<dyn Error>> {
28    pel::init();
29
30    // Dynamics models require planetary constants and ephemerides to be defined.
31    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
32    // This will automatically download the DE440s planetary ephemeris,
33    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
34    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
35    // planetary constants kernels.
36    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
37    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
38    // references to many functions.
39    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
40    // Fetch the EME2000 frame from the Almabac
41    let eme2k = almanac.frame_info(EARTH_J2000).unwrap();
42    // Define the orbit epoch
43    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
44
45    // Build the spacecraft itself.
46    // Using slide 6 of https://aerospace.org/sites/default/files/2018-11/Davis-Mayberry_HPSEP_11212018.pdf
47    // for the "next gen" SEP characteristics.
48
49    // GTO start
50    let orbit = Orbit::keplerian(24505.9, 0.725, 7.05, 0.0, 0.0, 0.0, epoch, eme2k);
51
52    let sc = Spacecraft::builder()
53        .orbit(orbit)
54        .mass(Mass::from_dry_and_prop_masses(1000.0, 1000.0)) // 1000 kg of dry mass and prop, totalling 2.0 tons
55        .srp(SRPData::from_area(3.0 * 6.0)) // Assuming 1 kW/m^2 or 18 kW, giving a margin of 4.35 kW for on-propulsion consumption
56        .thruster(Thruster {
57            // "NEXT-STEP" row in Table 2
58            isp_s: 4435.0,
59            thrust_N: 0.472,
60        })
61        .mode(GuidanceMode::Thrust) // Start thrusting immediately.
62        .build();
63
64    let prop_time = 180.0 * Unit::Day;
65
66    // Define the guidance law -- we're just using a Ruggiero controller as demonstrated in AAS-2004-5089.
67    let objectives = &[
68        Objective::within_tolerance(
69            StateParameter::Element(OrbitalElement::SemiMajorAxis),
70            42_165.0,
71            20.0,
72        ),
73        Objective::within_tolerance(
74            StateParameter::Element(OrbitalElement::Eccentricity),
75            0.001,
76            5e-5,
77        ),
78        Objective::within_tolerance(
79            StateParameter::Element(OrbitalElement::Inclination),
80            0.05,
81            1e-2,
82        ),
83    ];
84
85    // Ensure that we only thrust if we have more than 20% illumination.
86    let ruggiero_ctrl = Ruggiero::from_max_eclipse(objectives, sc, 0.2).unwrap();
87    println!("{ruggiero_ctrl}");
88
89    // Define the high fidelity dynamics
90
91    // Set up the spacecraft dynamics.
92
93    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
94    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
95    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
96
97    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
98    // We're using the JGM3 model here, which is the default in GMAT.
99    let mut jgm3_meta = MetaFile {
100        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
101        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
102    };
103    // And let's download it if we don't have it yet.
104    jgm3_meta.process(true)?;
105
106    // Build the spherical harmonics.
107    // The harmonics must be computed in the body fixed frame.
108    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
109    let harmonics = GravityField::new(
110        GravityFieldData::from_cof(&jgm3_meta.uri, 8, 8, almanac.frame_info(IAU_EARTH_FRAME)?)
111            .unwrap(),
112    );
113
114    // Include the spherical harmonics into the orbital dynamics.
115    orbital_dyn.accel_models.push(harmonics);
116
117    // We define the solar radiation pressure, using the default solar flux and accounting only
118    // for the eclipsing caused by the Earth.
119    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
120
121    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
122    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
123    let sc_dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn)
124        .with_guidance_law(ruggiero_ctrl.clone());
125
126    println!("{orbit:x}");
127
128    // We specify a minimum step in the propagator because the Ruggiero control would otherwise drive this step very low.
129    let (final_state, traj) = Propagator::rk89(
130        sc_dynamics.clone(),
131        IntegratorOptions::builder()
132            .min_step(10.0_f64.seconds())
133            .error_ctrl(ErrorControl::RSSCartesianStep)
134            .build(),
135    )
136    .with(sc, almanac.clone())
137    .for_duration_with_traj(prop_time)?;
138
139    let prop_usage = sc.mass.prop_mass_kg - final_state.mass.prop_mass_kg;
140    println!("{:x}", final_state.orbit);
141    println!("prop usage: {prop_usage:.3} kg");
142
143    // Finally, export the results for analysis, including the penumbra percentage throughout the orbit raise.
144    traj.to_parquet("./03_geo_raise.parquet", ExportCfg::default())?;
145
146    for status_line in ruggiero_ctrl.status(&final_state) {
147        println!("{status_line}");
148    }
149
150    ruggiero_ctrl
151        .achieved(&final_state)
152        .expect("objective not achieved");
153
154    Ok(())
155}
nyx-core/examples/01_orbit_prop/main.rs (line 133)
30fn main() -> Result<(), Box<dyn Error>> {
31    pel::init();
32    // Dynamics models require planetary constants and ephemerides to be defined.
33    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
34    // This will automatically download the DE440s planetary ephemeris,
35    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
36    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
37    // planetary constants kernels.
38    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
39    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
40    // references to many functions.
41    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
42    // Define the orbit epoch
43    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
44
45    // Define the orbit.
46    // First we need to fetch the Earth J2000 from information from the Almanac.
47    // This allows the frame to include the gravitational parameters and the shape of the Earth,
48    // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
49    // by loading a different set of planetary constants.
50    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
51
52    let orbit =
53        Orbit::try_keplerian_altitude(300.0, 0.015, 68.5, 65.2, 75.0, 0.0, epoch, earth_j2000)?;
54    // Print in in Keplerian form.
55    println!("{orbit:x}");
56
57    // There are two ways to propagate an orbit. We can make a quick approximation assuming only two-body
58    // motion. This is a useful first order approximation but it isn't used in real-world applications.
59
60    // This approach is a feature of ANISE.
61    let future_orbit_tb = orbit.at_epoch(epoch + Unit::Day * 3)?;
62    println!("{future_orbit_tb:x}");
63
64    // Two body propagation relies solely on Kepler's laws, so only the true anomaly will change.
65    println!(
66        "SMA changed by {:.3e} km",
67        orbit.sma_km()? - future_orbit_tb.sma_km()?
68    );
69    println!(
70        "ECC changed by {:.3e}",
71        orbit.ecc()? - future_orbit_tb.ecc()?
72    );
73    println!(
74        "INC changed by {:.3e} deg",
75        orbit.inc_deg()? - future_orbit_tb.inc_deg()?
76    );
77    println!(
78        "RAAN changed by {:.3e} deg",
79        orbit.raan_deg()? - future_orbit_tb.raan_deg()?
80    );
81    println!(
82        "AOP changed by {:.3e} deg",
83        orbit.aop_deg()? - future_orbit_tb.aop_deg()?
84    );
85    println!(
86        "TA changed by {:.3} deg",
87        orbit.ta_deg()? - future_orbit_tb.ta_deg()?
88    );
89
90    // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
91    // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
92    // models such as solar radiation pressure.
93
94    // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
95    let sc = Spacecraft::builder()
96        .orbit(orbit)
97        .mass(Mass::from_dry_mass(9.60))
98        .srp(SRPData {
99            area_m2: 10e-4,
100            coeff_reflectivity: 1.1,
101        })
102        .build();
103    println!("{sc:x}");
104
105    // Set up the spacecraft dynamics.
106
107    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
108    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
109    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
110
111    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
112    // We're using the JGM3 model here, which is the default in GMAT.
113    let mut jgm3_meta = MetaFile {
114        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
115        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
116    };
117    // And let's download it if we don't have it yet.
118    jgm3_meta.process(true)?;
119
120    // Build the spherical harmonics.
121    // The harmonics must be computed in the body fixed frame.
122    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
123    let harmonics_21x21 = GravityField::new(
124        GravityFieldData::from_cof(&jgm3_meta.uri, 21, 21, almanac.frame_info(IAU_EARTH_FRAME)?)
125            .unwrap(),
126    );
127
128    // Include the spherical harmonics into the orbital dynamics.
129    orbital_dyn.accel_models.push(harmonics_21x21);
130
131    // We define the solar radiation pressure, using the default solar flux and accounting only
132    // for the eclipsing caused by the Earth.
133    let srp_dyn = SolarPressure::default_flux(EARTH_J2000, &almanac)?;
134
135    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
136    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
137    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
138
139    println!("{dynamics}");
140
141    // Finally, let's propagate this orbit to the same epoch as above.
142    // The first returned value is the spacecraft state at the final epoch.
143    // The second value is the full trajectory where the step size is variable step used by the propagator.
144    let (future_sc, trajectory) = Propagator::default(dynamics)
145        .with(sc, almanac.clone())
146        .until_epoch_with_traj(future_orbit_tb.epoch)?;
147
148    println!("=== High fidelity propagation ===");
149    println!(
150        "SMA changed by {:.3} km",
151        orbit.sma_km()? - future_sc.orbit.sma_km()?
152    );
153    println!(
154        "ECC changed by {:.6}",
155        orbit.ecc()? - future_sc.orbit.ecc()?
156    );
157    println!(
158        "INC changed by {:.3e} deg",
159        orbit.inc_deg()? - future_sc.orbit.inc_deg()?
160    );
161    println!(
162        "RAAN changed by {:.3} deg",
163        orbit.raan_deg()? - future_sc.orbit.raan_deg()?
164    );
165    println!(
166        "AOP changed by {:.3} deg",
167        orbit.aop_deg()? - future_sc.orbit.aop_deg()?
168    );
169    println!(
170        "TA changed by {:.3} deg",
171        orbit.ta_deg()? - future_sc.orbit.ta_deg()?
172    );
173
174    // We also have access to the full trajectory throughout the propagation.
175    println!("{trajectory}");
176
177    // With the trajectory, let's build a few data products.
178
179    // 1. Export the trajectory as a CCSDS OEM version 2.0 file and as a parquet file, which includes the Keplerian orbital elements.
180
181    trajectory.to_oem_file(
182        "./01_cubesat_hf_prop.oem",
183        "CUBESAT-ID".to_string(),
184        Some("Nyx Space".to_string()),
185        Some("CUBESAT".to_string()),
186        ExportCfg::builder().step(Unit::Minute * 2).build(),
187    )?;
188
189    trajectory.to_parquet_with_cfg(
190        "./01_cubesat_hf_prop.parquet",
191        ExportCfg::builder().step(Unit::Minute * 2).build(),
192    )?;
193
194    // 2. Compare the difference in the radial-intrack-crosstrack frame between the high fidelity
195    // and Keplerian propagation. The RIC frame is commonly used to compute the difference in position
196    // and velocity of different spacecraft.
197    // 3. Compute the azimuth, elevation, range, and range-rate data of that spacecraft as seen from Boulder, CO, USA.
198
199    let boulder_station = GroundStation::from_point(
200        "Boulder, CO, USA".to_string(),
201        40.014984,   // latitude in degrees
202        -105.270546, // longitude in degrees
203        1.6550,      // altitude in kilometers
204        almanac.frame_info(IAU_EARTH_FRAME)?,
205    );
206
207    // We iterate over the trajectory, grabbing a state every two minutes.
208    let mut offset_s = vec![];
209    let mut epoch_str = vec![];
210    let mut ric_x_km = vec![];
211    let mut ric_y_km = vec![];
212    let mut ric_z_km = vec![];
213    let mut ric_vx_km_s = vec![];
214    let mut ric_vy_km_s = vec![];
215    let mut ric_vz_km_s = vec![];
216
217    let mut azimuth_deg = vec![];
218    let mut elevation_deg = vec![];
219    let mut range_km = vec![];
220    let mut range_rate_km_s = vec![];
221    for state in trajectory.every(Unit::Minute * 2) {
222        // Try to compute the Keplerian/two body state just in time.
223        // This method occasionally fails to converge on an appropriate true anomaly
224        // from the mean anomaly. If that happens, we just skip this state.
225        // The high fidelity and Keplerian states diverge continuously, and we're curious
226        // about the divergence in this quick analysis.
227        let this_epoch = state.epoch();
228        match orbit.at_epoch(this_epoch) {
229            Ok(tb_then) => {
230                offset_s.push((this_epoch - orbit.epoch).to_seconds());
231                epoch_str.push(format!("{this_epoch}"));
232                // Compute the two body state just in time.
233                let ric = state.orbit.ric_difference(&tb_then)?;
234                ric_x_km.push(ric.radius_km.x);
235                ric_y_km.push(ric.radius_km.y);
236                ric_z_km.push(ric.radius_km.z);
237                ric_vx_km_s.push(ric.velocity_km_s.x);
238                ric_vy_km_s.push(ric.velocity_km_s.y);
239                ric_vz_km_s.push(ric.velocity_km_s.z);
240
241                // Compute the AER data for each state.
242                let aer = almanac.azimuth_elevation_range_sez(
243                    state.orbit,
244                    boulder_station.to_orbit(this_epoch, &almanac)?,
245                    None,
246                    None,
247                )?;
248                azimuth_deg.push(aer.azimuth_deg);
249                elevation_deg.push(aer.elevation_deg);
250                range_km.push(aer.range_km);
251                range_rate_km_s.push(aer.range_rate_km_s);
252            }
253            Err(e) => warn!("{} {e}", state.epoch()),
254        };
255    }
256
257    // Build the data frames.
258    let ric_df = df!(
259        "Offset (s)" => offset_s.clone(),
260        "Epoch" => epoch_str.clone(),
261        "RIC X (km)" => ric_x_km,
262        "RIC Y (km)" => ric_y_km,
263        "RIC Z (km)" => ric_z_km,
264        "RIC VX (km/s)" => ric_vx_km_s,
265        "RIC VY (km/s)" => ric_vy_km_s,
266        "RIC VZ (km/s)" => ric_vz_km_s,
267    )?;
268
269    println!("RIC difference at start\n{}", ric_df.head(Some(10)));
270    println!("RIC difference at end\n{}", ric_df.tail(Some(10)));
271
272    let aer_df = df!(
273        "Offset (s)" => offset_s.clone(),
274        "Epoch" => epoch_str.clone(),
275        "azimuth (deg)" => azimuth_deg,
276        "elevation (deg)" => elevation_deg,
277        "range (km)" => range_km,
278        "range rate (km/s)" => range_rate_km_s,
279    )?;
280
281    // Finally, let's see when the spacecraft is visible, assuming 15 degrees minimum elevation.
282    let mask = aer_df
283        .column("elevation (deg)")?
284        .gt(&Column::Scalar(ScalarColumn::new(
285            "elevation mask (deg)".into(),
286            Scalar::new(DataType::Float64, AnyValue::Float64(15.0)),
287            offset_s.len(),
288        )))?;
289    let cubesat_visible = aer_df.filter(&mask)?;
290
291    println!("{cubesat_visible}");
292
293    Ok(())
294}
pub fn default_no_estimation( shadow_bodies: Vec<Frame>, almanac: &Almanac, ) -> Result<Arc<Self>, DynamicsError>

Accounts for the shadowing of only one body and will set the solar flux at 1 AU to: Phi = 1367.0

pub fn with_flux( flux_w_m2: f64, shadow_bodies: Vec<Frame>, almanac: &Almanac, ) -> Result<Arc<Self>, DynamicsError>

Must provide the flux in W/m^2

pub fn new( shadow_bodies: Vec<Frame>, almanac: &Almanac, ) -> Result<Arc<Self>, DynamicsError>

Solar radiation pressure force model accounting for the provided shadow bodies.

Examples found in repository?
nyx-core/examples/02_jwst_covar_monte_carlo/main.rs (line 100)
26fn main() -> Result<(), Box<dyn Error>> {
27    pel::init();
28    // Dynamics models require planetary constants and ephemerides to be defined.
29    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
31
32    // Download the regularly update of the James Webb Space Telescope reconstucted (or definitive) ephemeris.
33    // Refer to https://naif.jpl.nasa.gov/pub/naif/JWST/kernels/spk/aareadme.txt for details.
34    let mut latest_jwst_ephem = MetaFile {
35        uri: "https://naif.jpl.nasa.gov/pub/naif/JWST/kernels/spk/jwst_rec.bsp".to_string(),
36        crc32: None,
37    };
38    latest_jwst_ephem.process(true)?;
39
40    // Load this ephem in the general Almanac we're using for this analysis.
41    let almanac = Arc::new(
42        MetaAlmanac::latest()
43            .map_err(Box::new)?
44            .load_from_metafile(latest_jwst_ephem, true)?,
45    );
46
47    // By loading this ephemeris file in the ANISE GUI or ANISE CLI, we can find the NAIF ID of the JWST
48    // in the BSP. We need this ID in order to query the ephemeris.
49    const JWST_NAIF_ID: i32 = -170;
50    // Let's build a frame in the J2000 orientation centered on the JWST.
51    const JWST_J2000: Frame = Frame::from_ephem_j2000(JWST_NAIF_ID);
52
53    // Since the ephemeris file is updated regularly, we'll just grab the latest state in the ephem.
54    let (earliest_epoch, latest_epoch) = almanac.spk_domain(JWST_NAIF_ID)?;
55    println!("JWST defined from {earliest_epoch} to {latest_epoch}");
56    // Fetch the state, printing it in the Earth J2000 frame.
57    let jwst_orbit = almanac.transform(JWST_J2000, EARTH_J2000, latest_epoch, None)?;
58    println!("{jwst_orbit:x}");
59
60    // Build the spacecraft
61    // SRP area assumed to be the full sunshield and mass if 6200.0 kg, c.f. https://webb.nasa.gov/content/about/faqs/facts.html
62    // SRP Coefficient of reflectivity assumed to be that of Kapton, i.e. 2 - 0.44 = 1.56, table 1 from https://amostech.com/TechnicalPapers/2018/Poster/Bengtson.pdf
63    let jwst = Spacecraft::builder()
64        .orbit(jwst_orbit)
65        .srp(SRPData {
66            area_m2: 21.197 * 14.162,
67            coeff_reflectivity: 1.56,
68        })
69        .mass(Mass::from_dry_mass(6200.0))
70        .build();
71
72    // Build up the spacecraft uncertainty builder.
73    // We can use the spacecraft uncertainty structure to build this up.
74    // We start by specifying the nominal state (as defined above), then the uncertainty in position and velocity
75    // in the RIC frame. We could also specify the Cr, Cd, and mass uncertainties, but these aren't accounted for until
76    // Nyx can also estimate the deviation of the spacecraft parameters.
77    let jwst_uncertainty = SpacecraftUncertainty::builder()
78        .nominal(jwst)
79        .frame(LocalFrame::RIC)
80        .x_km(0.5)
81        .y_km(0.3)
82        .z_km(1.5)
83        .vx_km_s(1e-4)
84        .vy_km_s(0.6e-3)
85        .vz_km_s(3e-3)
86        .build();
87
88    println!("{jwst_uncertainty}");
89
90    // Build the Kalman filter estimate.
91    // Note that we could have used the KfEstimate structure directly (as seen throughout the OD integration tests)
92    // but this approach requires quite a bit more boilerplate code.
93    let jwst_estimate = jwst_uncertainty.to_estimate()?;
94
95    // Set up the spacecraft dynamics.
96    // We'll use the point masses of the Earth, Sun, Jupiter (barycenter, because it's in the DE440), and the Moon.
97    // We'll also enable solar radiation pressure since the James Webb has a huge and highly reflective sun shield.
98
99    let orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN, JUPITER_BARYCENTER]);
100    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
101
102    // Finalize setting up the dynamics.
103    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
104
105    // Build the propagator set up to use for the whole analysis.
106    let setup = Propagator::default(dynamics);
107
108    // All of the analysis will use this duration.
109    let prediction_duration = 6.5 * Unit::Day;
110
111    // === Covariance mapping ===
112    // For the covariance mapping / prediction, we'll use the common orbit determination approach.
113    // This is done by setting up a spacecraft Kalman filter OD process, and predicting for the analysis duration.
114
115    // Build the propagation instance for the OD process.
116    let odp = SpacecraftKalmanOD::new(
117        setup.clone(),
118        KalmanVariant::DeviationTracking,
119        None,
120        BTreeMap::new(),
121        almanac.clone(),
122    );
123
124    // The prediction step is 1 minute by default, configured in the OD process, i.e. how often we want to know the covariance.
125    assert_eq!(odp.max_step, 1_i64.minutes());
126    // Finally, predict, and export the trajectory with covariance to a parquet file.
127    let od_sol = odp.predict_for(jwst_estimate, prediction_duration)?;
128    od_sol.to_parquet("./02_jwst_covar_map.parquet", ExportCfg::default())?;
129
130    // === Monte Carlo framework ===
131    // Nyx comes with a complete multi-threaded Monte Carlo frame. It's blazing fast.
132
133    let my_mc = MonteCarlo::new(
134        jwst, // Nominal state
135        jwst_estimate.to_random_variable()?,
136        "02_jwst".to_string(), // Scenario name
137        None, // No specific seed specified, so one will be drawn from the computer's entropy.
138    );
139
140    let num_runs = 5_000;
141    let rslts = my_mc.run_until_epoch(
142        setup,
143        almanac.clone(),
144        jwst.epoch() + prediction_duration,
145        num_runs,
146    );
147
148    assert_eq!(rslts.runs.len(), num_runs);
149    // Finally, export these results, computing the eclipse percentage for all of these results.
150
151    rslts.to_parquet("02_jwst_monte_carlo.parquet", ExportCfg::default())?;
152
153    Ok(())
154}
More examples
Hide additional examples
nyx-core/examples/06_lunar_orbit_determination/main.rs (line 118)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let data_folder: PathBuf = [
46        env!("CARGO_MANIFEST_DIR"),
47        "examples",
48        "06_lunar_orbit_determination",
49    ]
50    .iter()
51    .collect();
52
53    let meta = data_folder.join("metaalmanac.dhall");
54
55    // Load this ephem in the general Almanac we're using for this analysis.
56    let almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57        .map_err(Box::new)?
58        .process(true)
59        .map_err(Box::new)?;
60
61    // Lock the almanac (an Arc is a read only structure).
62    let almanac = Arc::new(almanac);
63
64    // Build a nominal trajectory
65    // TODO: Switch this to a sequence once the OD over a spacecraft sequence is implemented.
66
67    let epoch = Epoch::from_gregorian_utc_at_noon(2024, 2, 29);
68    let moon_j2000 = almanac.frame_info(MOON_J2000)?;
69
70    // To build the trajectory we need to provide a spacecraft template.
71    let orbiter = Spacecraft::builder()
72        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0))
73        .srp(SRPData {
74            area_m2: 3.9 * 2.7,
75            coeff_reflectivity: 0.96,
76        })
77        .orbit(Orbit::try_keplerian_altitude(
78            150.0, 0.00212, 33.6, 45.0, 45.0, 0.0, epoch, moon_j2000,
79        )?) // Setting a zero orbit here because it's just a template
80        .build();
81
82    // ========================== //
83    // === BUILD NOMINAL TRAJ === //
84    // ========================== //
85
86    // Set up the spacecraft dynamics.
87
88    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
89    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
90    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
91
92    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
93    // We're using the GRAIL JGGRX model.
94    let mut jggrx_meta = MetaFile {
95        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
96        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
97    };
98    // And let's download it if we don't have it yet.
99    jggrx_meta.process(true)?;
100
101    // Build the spherical harmonics.
102    // The harmonics must be computed in the body fixed frame.
103    // We're using the long term prediction of the Moon principal axes frame.
104    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
105    let sph_harmonics = GravityField::new(GravityFieldData::from_shadr(
106        &jggrx_meta.uri,
107        80,
108        80,
109        almanac.frame_info(moon_pa_frame)?,
110    )?);
111
112    // Include the spherical harmonics into the orbital dynamics.
113    orbital_dyn.accel_models.push(sph_harmonics);
114
115    // We define the solar radiation pressure, using the default solar flux and accounting only
116    // for the eclipsing caused by the Earth and Moon.
117    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
118    let srp_dyn = SolarPressure::new(vec![MOON_J2000], &almanac)?;
119
120    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
121    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
122    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
123
124    println!("{dynamics}");
125
126    let setup = Propagator::rk89(dynamics.clone(), IntegratorOptions::default());
127
128    let truth_traj = setup
129        .with(orbiter, almanac.clone())
130        .for_duration_with_traj(Unit::Day * 2)?
131        .1;
132
133    // ==================== //
134    // === OD SIMULATOR === //
135    // ==================== //
136
137    // Load the Deep Space Network ground stations.
138    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
139    let ground_station_file = data_folder.join("dsn-network.yaml");
140    let devices = GroundStation::load_named(ground_station_file)?;
141
142    let proc_devices = devices.clone();
143
144    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
145    // Nyx can build a tracking schedule for you based on the first station with access.
146    let configs: BTreeMap<String, TrkConfig> =
147        TrkConfig::load_named(data_folder.join("tracking-cfg.yaml"))?;
148
149    // Build the tracking arc simulation to generate a "standard measurement".
150    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
151        devices.clone(),
152        truth_traj.clone(),
153        configs,
154        123, // Set a seed for reproducibility
155    )?;
156
157    trk.build_schedule(&almanac)?;
158    let arc = trk.generate_measurements(&almanac)?;
159    // Save the simulated tracking data
160    arc.to_parquet_simple("./data/04_output/06_lunar_simulated_tracking.parquet")?;
161
162    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
163    println!("{arc}");
164
165    // Now that we have simulated measurements, we'll run the orbit determination.
166
167    // ===================== //
168    // === OD ESTIMATION === //
169    // ===================== //
170
171    let sc = SpacecraftUncertainty::builder()
172        .nominal(orbiter)
173        .frame(LocalFrame::RIC)
174        .x_km(0.5)
175        .y_km(0.5)
176        .z_km(0.5)
177        .vx_km_s(5e-3)
178        .vy_km_s(5e-3)
179        .vz_km_s(5e-3)
180        .build();
181
182    // Build the filter initial estimate, which we will reuse in the filter.
183    let initial_estimate = sc.to_estimate()?;
184
185    println!("== FILTER STATE ==\n{orbiter:x}\n{initial_estimate}");
186
187    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
188    let process_noise = ProcessNoise3D::from_velocity_km_s(
189        &[1e-14, 1e-14, 1e-14],
190        1 * Unit::Hour,
191        10 * Unit::Minute,
192        None,
193    );
194
195    println!("{process_noise}");
196
197    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
198    let odp = SpacecraftKalmanScalarOD::new(
199        setup,
200        KalmanVariant::ReferenceUpdate,
201        Some(SigmaRejection::default()),
202        proc_devices,
203        almanac.clone(),
204    )
205    .with_process_noise(process_noise);
206
207    let od_sol = odp.process_arc(initial_estimate, &arc)?;
208
209    let final_est = od_sol.estimates.last().unwrap();
210
211    println!("{final_est}");
212
213    let ric_err = truth_traj
214        .at(final_est.epoch())?
215        .orbit
216        .ric_difference(&final_est.orbital_state())?;
217    println!("== RIC at end ==");
218    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
219    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
220
221    println!(
222        "Num residuals rejected: #{}",
223        od_sol.rejected_residuals().len()
224    );
225    println!(
226        "Percentage within +/-3: {}",
227        od_sol.residual_ratio_within_threshold(3.0).unwrap()
228    );
229    println!("Whitened residuals normal? {}", od_sol.is_normal(None)?);
230    println!("NIS consistency: {}", od_sol.nis_consistency(None)?);
231
232    od_sol.to_parquet(
233        "./data/04_output/06_lunar_od_results.parquet",
234        ExportCfg::default(),
235    )?;
236
237    let od_trajectory = od_sol.to_traj()?;
238    // Build the RIC difference.
239    od_trajectory.ric_diff_to_parquet(
240        &truth_traj,
241        "./data/04_output/06_lunar_od_truth_error.parquet",
242        ExportCfg::default(),
243    )?;
244
245    Ok(())
246}
nyx-core/examples/03_geo_analysis/drift.rs (line 100)
26fn main() -> Result<(), Box<dyn Error>> {
27    pel::init();
28    // Dynamics models require planetary constants and ephemerides to be defined.
29    // Let's start by grabbing those by using ANISE's latest MetaAlmanac.
30    // This will automatically download the DE440s planetary ephemeris,
31    // the daily-updated Earth Orientation Parameters, the high fidelity Moon orientation
32    // parameters (for the Moon Mean Earth and Moon Principal Axes frames), and the PCK11
33    // planetary constants kernels.
34    // For details, refer to https://github.com/nyx-space/anise/blob/master/data/latest.dhall.
35    // Note that we place the Almanac into an Arc so we can clone it cheaply and provide read-only
36    // references to many functions.
37    let almanac = Arc::new(MetaAlmanac::latest().map_err(Box::new)?);
38    // Define the orbit epoch
39    let epoch = Epoch::from_gregorian_utc_hms(2024, 2, 29, 12, 13, 14);
40
41    // Define the orbit.
42    // First we need to fetch the Earth J2000 from information from the Almanac.
43    // This allows the frame to include the gravitational parameters and the shape of the Earth,
44    // defined as a tri-axial ellipoid. Note that this shape can be changed manually or in the Almanac
45    // by loading a different set of planetary constants.
46    let earth_j2000 = almanac.frame_info(EARTH_J2000)?;
47
48    // Placing this GEO bird just above Colorado.
49    // In theory, the eccentricity is zero, but in practice, it's about 1e-5 to 1e-6 at best.
50    let orbit = Orbit::try_keplerian(42164.0, 1e-5, 0., 163.0, 75.0, 0.0, epoch, earth_j2000)?;
51    // Print in in Keplerian form.
52    println!("{orbit:x}");
53
54    let state_bf = almanac.transform_to(orbit, IAU_EARTH_FRAME, None)?;
55    let (orig_lat_deg, orig_long_deg, orig_alt_km) = state_bf.latlongalt()?;
56
57    // Nyx is used for high fidelity propagation, not Keplerian propagation as above.
58    // Nyx only propagates Spacecraft at the moment, which allows it to account for acceleration
59    // models such as solar radiation pressure.
60
61    // Let's build a cubesat sized spacecraft, with an SRP area of 10 cm^2 and a mass of 9.6 kg.
62    let sc = Spacecraft::builder()
63        .orbit(orbit)
64        .mass(Mass::from_dry_mass(9.60))
65        .srp(SRPData {
66            area_m2: 10e-4,
67            coeff_reflectivity: 1.1,
68        })
69        .build();
70    println!("{sc:x}");
71
72    // Set up the spacecraft dynamics.
73
74    // Specify that the orbital dynamics must account for the graviational pull of the Moon and the Sun.
75    // The gravity of the Earth will also be accounted for since the spaceraft in an Earth orbit.
76    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![MOON, SUN]);
77
78    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
79    // We're using the JGM3 model here, which is the default in GMAT.
80    let mut jgm3_meta = MetaFile {
81        uri: "http://public-data.nyxspace.com/nyx/models/JGM3.cof.gz".to_string(),
82        crc32: Some(0xF446F027), // Specifying the CRC32 avoids redownloading it if it's cached.
83    };
84    // And let's download it if we don't have it yet.
85    jgm3_meta.process(true)?;
86
87    // Build the spherical harmonics.
88    // The harmonics must be computed in the body fixed frame.
89    // We're using the long term prediction of the Earth centered Earth fixed frame, IAU Earth.
90    let harmonics_21x21 = GravityField::new(
91        GravityFieldData::from_cof(&jgm3_meta.uri, 21, 21, almanac.frame_info(IAU_EARTH_FRAME)?)
92            .unwrap(),
93    );
94
95    // Include the spherical harmonics into the orbital dynamics.
96    orbital_dyn.accel_models.push(harmonics_21x21);
97
98    // We define the solar radiation pressure, using the default solar flux and accounting only
99    // for the eclipsing caused by the Earth and Moon.
100    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
101
102    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
103    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
104    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
105
106    println!("{dynamics}");
107
108    // Finally, let's propagate this orbit to the same epoch as above.
109    // The first returned value is the spacecraft state at the final epoch.
110    // The second value is the full trajectory where the step size is variable step used by the propagator.
111    let (future_sc, trajectory) = Propagator::default(dynamics)
112        .with(sc, almanac.clone())
113        .until_epoch_with_traj(epoch + Unit::Century * 0.03)?;
114
115    println!("=== High fidelity propagation ===");
116    println!(
117        "SMA changed by {:.3} km",
118        orbit.sma_km()? - future_sc.orbit.sma_km()?
119    );
120    println!(
121        "ECC changed by {:.6}",
122        orbit.ecc()? - future_sc.orbit.ecc()?
123    );
124    println!(
125        "INC changed by {:.3e} deg",
126        orbit.inc_deg()? - future_sc.orbit.inc_deg()?
127    );
128    println!(
129        "RAAN changed by {:.3} deg",
130        orbit.raan_deg()? - future_sc.orbit.raan_deg()?
131    );
132    println!(
133        "AOP changed by {:.3} deg",
134        orbit.aop_deg()? - future_sc.orbit.aop_deg()?
135    );
136    println!(
137        "TA changed by {:.3} deg",
138        orbit.ta_deg()? - future_sc.orbit.ta_deg()?
139    );
140
141    // We also have access to the full trajectory throughout the propagation.
142    println!("{trajectory}");
143
144    println!("Spacecraft params after 3 years without active control:\n{future_sc:x}");
145
146    // With the trajectory, let's build a few data products.
147
148    // 1. Export the trajectory as a parquet file, which includes the Keplerian orbital elements.
149
150    let analysis_step = Unit::Minute * 5;
151
152    trajectory.to_parquet(
153        "./03_geo_hf_prop.parquet",
154        ExportCfg::builder().step(analysis_step).build(),
155    )?;
156
157    // 2. Compute the latitude, longitude, and altitude throughout the trajectory by rotating the spacecraft position into the Earth body fixed frame.
158
159    // We iterate over the trajectory, grabbing a state every two minutes.
160    let mut offset_s = vec![];
161    let mut epoch_str = vec![];
162    let mut longitude_deg = vec![];
163    let mut latitude_deg = vec![];
164    let mut altitude_km = vec![];
165
166    for state in trajectory.every(analysis_step) {
167        // Convert the GEO bird state into the body fixed frame, and keep track of its latitude, longitude, and altitude.
168        // These define the GEO stationkeeping box.
169
170        let this_epoch = state.epoch();
171
172        offset_s.push((this_epoch - orbit.epoch).to_seconds());
173        epoch_str.push(this_epoch.to_isoformat());
174
175        let state_bf = almanac.transform_to(state.orbit, IAU_EARTH_FRAME, None)?;
176        let (lat_deg, long_deg, alt_km) = state_bf.latlongalt()?;
177        longitude_deg.push(long_deg);
178        latitude_deg.push(lat_deg);
179        altitude_km.push(alt_km);
180    }
181
182    println!(
183        "Longitude changed by {:.3} deg -- Box is 0.1 deg E-W",
184        orig_long_deg - longitude_deg.last().unwrap()
185    );
186
187    println!(
188        "Latitude changed by {:.3} deg -- Box is 0.05 deg N-S",
189        orig_lat_deg - latitude_deg.last().unwrap()
190    );
191
192    println!(
193        "Altitude changed by {:.3} km -- Box is 30 km",
194        orig_alt_km - altitude_km.last().unwrap()
195    );
196
197    // Build the station keeping data frame.
198    let mut sk_df = df!(
199        "Offset (s)" => offset_s.clone(),
200        "Epoch (UTC)" => epoch_str.clone(),
201        "Longitude E-W (deg)" => longitude_deg,
202        "Latitude N-S (deg)" => latitude_deg,
203        "Altitude (km)" => altitude_km,
204
205    )?;
206
207    // Create a file to write the Parquet to
208    let file = File::create("./03_geo_lla.parquet").expect("Could not create file");
209
210    // Create a ParquetWriter and write the DataFrame to the file
211    ParquetWriter::new(file).finish(&mut sk_df)?;
212
213    Ok(())
214}
nyx-core/examples/04_lro_od/main.rs (line 149)
35fn main() -> Result<(), Box<dyn Error>> {
36    pel::init();
37
38    // ====================== //
39    // === ALMANAC SET UP === //
40    // ====================== //
41
42    // Dynamics models require planetary constants and ephemerides to be defined.
43    // Let's start by grabbing those by using ANISE's MetaAlmanac.
44
45    let output_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "../data", "04_output"]
46        .iter()
47        .collect();
48
49    let data_folder: PathBuf = [env!("CARGO_MANIFEST_DIR"), "examples", "04_lro_od"]
50        .iter()
51        .collect();
52
53    let meta = data_folder.join("lro-dynamics.dhall");
54
55    // Load this ephem in the general Almanac we're using for this analysis.
56    let mut almanac = MetaAlmanac::new(meta.to_string_lossy().as_ref())
57        .map_err(Box::new)?
58        .process(true)
59        .map_err(Box::new)?;
60
61    let mut moon_pc = almanac.get_planetary_data_from_id(MOON).unwrap();
62    moon_pc.mu_km3_s2 = 4902.74987;
63    almanac.set_planetary_data_from_id(MOON, moon_pc).unwrap();
64
65    let mut earth = almanac.get_planetary_data_from_id(EARTH).unwrap();
66    earth.mu_km3_s2 = 398600.436;
67    almanac.set_planetary_data_from_id(EARTH, earth).unwrap();
68
69    // Save this new kernel for reuse.
70    // In an operational context, this would be part of the "Lock" process, and should not change throughout the mission.
71    almanac
72        .planetary_data
73        .values()
74        .next()
75        .unwrap()
76        .save_as(&data_folder.join("lro-specific.pca"), true)?;
77
78    // Lock the almanac (an Arc is a read only structure).
79    let almanac = Arc::new(almanac);
80
81    // Orbit determination requires a Trajectory structure, which can be saved as parquet file.
82    // In our case, the trajectory comes from the BSP file, so we need to build a Trajectory from the almanac directly.
83    // To query the Almanac, we need to build the LRO frame in the J2000 orientation in our case.
84    // Inspecting the LRO BSP in the ANISE GUI shows us that NASA has assigned ID -85 to LRO.
85    let lro_frame = Frame::from_ephem_j2000(-85);
86
87    // To build the trajectory we need to provide a spacecraft template.
88    let sc_template = Spacecraft::builder()
89        .mass(Mass::from_dry_and_prop_masses(1018.0, 900.0)) // Launch masses
90        .srp(SRPData {
91            // SRP configuration is arbitrary, but we will be estimating it anyway.
92            area_m2: 3.9 * 2.7,
93            coeff_reflectivity: 0.96,
94        })
95        .orbit(Orbit::zero(MOON_J2000)) // Setting a zero orbit here because it's just a template
96        .build();
97    // Now we can build the trajectory from the BSP file.
98    // We'll arbitrarily set the tracking arc to 24 hours with a five second time step.
99    let traj_as_flown = Traj::from_bsp(
100        lro_frame,
101        MOON_J2000,
102        &almanac,
103        sc_template,
104        5.seconds(),
105        Some(Epoch::from_str("2024-01-01 00:00:00 UTC")?),
106        Some(Epoch::from_str("2024-01-02 00:00:00 UTC")?),
107        Aberration::LT,
108        Some("LRO".to_string()),
109    )?;
110
111    println!("{traj_as_flown}");
112
113    // ====================== //
114    // === MODEL MATCHING === //
115    // ====================== //
116
117    // Set up the spacecraft dynamics.
118
119    // Specify that the orbital dynamics must account for the graviational pull of the Earth and the Sun.
120    // The gravity of the Moon will also be accounted for since the spaceraft in a lunar orbit.
121    let mut orbital_dyn = OrbitalDynamics::point_masses(vec![EARTH, SUN, JUPITER_BARYCENTER]);
122
123    // We want to include the spherical harmonics, so let's download the gravitational data from the Nyx Cloud.
124    // We're using the GRAIL JGGRX model.
125    let mut jggrx_meta = MetaFile {
126        uri: "http://public-data.nyxspace.com/nyx/models/Luna_jggrx_1500e_sha.tab.gz".to_string(),
127        crc32: Some(0x6bcacda8), // Specifying the CRC32 avoids redownloading it if it's cached.
128    };
129    // And let's download it if we don't have it yet.
130    jggrx_meta.process(true)?;
131
132    // Build the spherical harmonics.
133    // The harmonics must be computed in the body fixed frame.
134    // We're using the long term prediction of the Moon principal axes frame.
135    let moon_pa_frame = MOON_PA_FRAME.with_orient(31008);
136    let sph_harmonics = GravityField::new(GravityFieldData::from_shadr(
137        &jggrx_meta.uri,
138        80,
139        80,
140        almanac.frame_info(moon_pa_frame)?,
141    )?);
142
143    // Include the spherical harmonics into the orbital dynamics.
144    orbital_dyn.accel_models.push(sph_harmonics);
145
146    // We define the solar radiation pressure, using the default solar flux and accounting only
147    // for the eclipsing caused by the Earth and Moon.
148    // Note that by default, enabling the SolarPressure model will also enable the estimation of the coefficient of reflectivity.
149    let srp_dyn = SolarPressure::new(vec![EARTH_J2000, MOON_J2000], &almanac)?;
150
151    // Finalize setting up the dynamics, specifying the force models (orbital_dyn) separately from the
152    // acceleration models (SRP in this case). Use `from_models` to specify multiple accel models.
153    let dynamics = SpacecraftDynamics::from_model(orbital_dyn, srp_dyn);
154
155    println!("{dynamics}");
156
157    // Now we can build the propagator.
158    let setup = Propagator::default_dp78(dynamics.clone());
159
160    // For reference, let's build the trajectory with Nyx's models from that LRO state.
161    let (sim_final, traj_as_sim) = setup
162        .with(*traj_as_flown.first(), almanac.clone())
163        .until_epoch_with_traj(traj_as_flown.last().epoch())?;
164
165    println!("SIM INIT:  {:x}", traj_as_flown.first());
166    println!("SIM FINAL: {sim_final:x}");
167    // Compute RIC difference between SIM and LRO ephem
168    let sim_lro_delta = sim_final
169        .orbit
170        .ric_difference(&traj_as_flown.last().orbit)?;
171    println!("{traj_as_sim}");
172    println!(
173        "SIM v LRO - RIC Position (m): {:.3}",
174        sim_lro_delta.radius_km * 1e3
175    );
176    println!(
177        "SIM v LRO - RIC Velocity (m/s): {:.3}",
178        sim_lro_delta.velocity_km_s * 1e3
179    );
180
181    traj_as_sim.ric_diff_to_parquet(
182        &traj_as_flown,
183        output_folder.join("./04_lro_sim_truth_error.parquet"),
184        ExportCfg::default(),
185    )?;
186
187    // ==================== //
188    // === OD SIMULATOR === //
189    // ==================== //
190
191    // After quite some time trying to exactly match the model, we still end up with an oscillatory difference on the order of 150 meters between the propagated state
192    // and the truth LRO state.
193
194    // Therefore, we will actually run an estimation from a dispersed LRO state.
195    // The sc_seed is the true LRO state from the BSP.
196    let sc_seed = *traj_as_flown.first();
197
198    // Load the Deep Space Network ground stations.
199    // Nyx allows you to build these at runtime but it's pretty static so we can just load them from YAML.
200    let ground_station_file: PathBuf = [
201        env!("CARGO_MANIFEST_DIR"),
202        "examples",
203        "04_lro_od",
204        "dsn-network.yaml",
205    ]
206    .iter()
207    .collect();
208
209    let devices = GroundStation::load_named(ground_station_file)?;
210
211    let mut proc_devices = devices.clone();
212
213    // Increase the noise in the devices to accept more measurements.
214    for gs in proc_devices.values_mut() {
215        if let Some(noise) = &mut gs
216            .stochastic_noises
217            .as_mut()
218            .unwrap()
219            .get_mut(&MeasurementType::Range)
220        {
221            *noise.white_noise.as_mut().unwrap() *= 3.0;
222        }
223    }
224
225    // Typical OD software requires that you specify your own tracking schedule or you'll have overlapping measurements.
226    // Nyx can build a tracking schedule for you based on the first station with access.
227    let trkconfg_yaml: PathBuf = [
228        env!("CARGO_MANIFEST_DIR"),
229        "examples",
230        "04_lro_od",
231        "tracking-cfg.yaml",
232    ]
233    .iter()
234    .collect();
235
236    let configs: BTreeMap<String, TrkConfig> = TrkConfig::load_named(trkconfg_yaml)?;
237
238    // Build the tracking arc simulation to generate a "standard measurement".
239    let mut trk = TrackingArcSim::<Spacecraft, GroundStation>::with_seed(
240        devices.clone(),
241        traj_as_flown.clone(),
242        configs,
243        123, // Set a seed for reproducibility
244    )?;
245
246    trk.build_schedule(&almanac)?;
247    let arc = trk.generate_measurements(&almanac)?;
248    // Save the simulated tracking data
249    arc.to_parquet_simple(output_folder.join("04_lro_simulated_tracking.parquet"))?;
250
251    // We'll note that in our case, we have continuous coverage of LRO when the vehicle is not behind the Moon.
252    println!("{arc}");
253
254    // Now that we have simulated measurements, we'll run the orbit determination.
255
256    // ===================== //
257    // === OD ESTIMATION === //
258    // ===================== //
259
260    let sc = SpacecraftUncertainty::builder()
261        .nominal(sc_seed)
262        .frame(LocalFrame::RIC)
263        .x_km(0.5)
264        .y_km(0.5)
265        .z_km(0.5)
266        .vx_km_s(5e-3)
267        .vy_km_s(5e-3)
268        .vz_km_s(5e-3)
269        .build();
270
271    // Build the filter initial estimate, which we will reuse in the filter.
272    let mut initial_estimate = sc.to_estimate()?;
273    initial_estimate.covar *= 3.0;
274
275    println!("== FILTER STATE ==\n{sc_seed:x}\n{initial_estimate}");
276
277    // Build the SNC in the Moon J2000 frame, specified as a velocity noise over time.
278    let process_noise = ProcessNoise3D::from_velocity_km_s(
279        &[1e-12, 1e-12, 1e-12],
280        1 * Unit::Hour,
281        10 * Unit::Minute,
282        None,
283    );
284
285    println!("{process_noise}");
286
287    // We'll set up the OD process to reject measurements whose residuals are move than 3 sigmas away from what we expect.
288    let odp = SpacecraftKalmanOD::new(
289        setup,
290        KalmanVariant::ReferenceUpdate,
291        Some(SigmaRejection::default()),
292        proc_devices,
293        almanac.clone(),
294    )
295    .with_process_noise(process_noise);
296
297    let od_sol = odp.process_arc(initial_estimate, &arc)?;
298
299    let final_est = od_sol.estimates.last().unwrap();
300
301    println!("{final_est}");
302
303    let ric_err = traj_as_flown
304        .at(final_est.epoch())?
305        .orbit
306        .ric_difference(&final_est.orbital_state())?;
307    println!("== RIC at end ==");
308    println!("RIC Position (m): {:.3}", ric_err.radius_km * 1e3);
309    println!("RIC Velocity (m/s): {:.3}", ric_err.velocity_km_s * 1e3);
310
311    println!(
312        "Num residuals rejected: #{}",
313        od_sol.rejected_residuals().len()
314    );
315    println!(
316        "Percentage within +/-3: {}",
317        od_sol.residual_ratio_within_threshold(3.0).unwrap()
318    );
319    println!("Ratios normal? {}", od_sol.is_normal(None).unwrap());
320
321    od_sol.to_parquet(
322        output_folder.join("04_lro_od_results.parquet"),
323        ExportCfg::default(),
324    )?;
325
326    // Create the ephemeris
327    let ephem = od_sol.to_ephemeris("LRO rebuilt".to_string());
328    let ephem_start = ephem.start_epoch().unwrap();
329    let ephem_end = ephem.end_epoch().unwrap();
330    // Check that the covariance is PSD throughout the ephemeris by interpolating it.
331    for epoch in TimeSeries::inclusive(ephem_start, ephem_end, Unit::Minute * 5) {
332        ephem
333            .covar_at(
334                epoch,
335                anise::ephemerides::ephemeris::LocalFrame::RIC,
336                &almanac,
337            )
338            .unwrap_or_else(|e| panic!("covar not PSD at {epoch}: {e}"));
339    }
340    // Export as BSP!
341    ephem
342        .write_spice_bsp(
343            -85,
344            output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap(),
345            None,
346        )
347        .expect("could not built BSP");
348    let new_almanac = Almanac::default()
349        .load(output_folder.join("04_lro_rebuilt.bsp").to_str().unwrap())
350        .unwrap();
351    new_almanac.describe(None, None, None, None, None, None, None, None);
352    let (spk_start, spk_end) = new_almanac.spk_domain(-85).unwrap();
353
354    assert!((ephem_start - spk_start).abs() < Unit::Microsecond * 1);
355    assert!((ephem_end - spk_end).abs() < Unit::Microsecond * 1);
356
357    // In our case, we have the truth trajectory from NASA.
358    // So we can compute the RIC state difference between the real LRO ephem and what we've just estimated.
359    // Export the OD trajectory first.
360    let od_trajectory = od_sol.to_traj()?;
361    // Build the RIC difference.
362    od_trajectory.ric_diff_to_parquet(
363        &traj_as_flown,
364        output_folder.join("04_lro_od_truth_error.parquet"),
365        ExportCfg::default(),
366    )?;
367
368    Ok(())
369}